from json import dumps
from urllib.parse import quote


def create_block_kit_builder_url(block_kit_like_object):
    """
    Takes block_kit_like_object like element, component, surfaces, and objects
    block_kit_like_object can be built (build()) or unbuilt, see example
    """
    block_kit_builder_url = "https://app.slack.com/block-kit-builder/#"
    try:
        block_dump = dumps(block_kit_like_object.build())
    except AttributeError as attribute_error:
        if str(attribute_error) == "'dict' object has no attribute 'build'":
            block_dump = dumps(block_kit_like_object)
        else:
            raise AttributeError
    except Exception as e:
        raise e from None
    final_url = block_kit_builder_url + block_dump
    final_url_safe = quote(final_url, safe="/:?=&#")
    msg = "Block kit builder example/validation:  \n\t" + final_url_safe
    print(msg)



from enum import Enum


class Style(str, Enum):
    primary = "primary"
    danger = "danger"


class TriggerActionsOn(str, Enum):
    on_enter_pressed = "on_enter_pressed"
    on_character_entered = "on_character_entered"


class Include(str, Enum):
    im = "im"
    mpim = "mpim"
    private = "private"
    public = "public"



from typing import List, Optional, Union

from pydantic import Field, root_validator

from blockkit.blocks import (
    Actions,
    Context,
    Divider,
    Header,
    ImageBlock,
    Input,
    Section,
)
from blockkit.components import Component
from blockkit.objects import PlainText
from blockkit.validators import validate_text_length, validator

__all__ = ["Home", "Message", "Modal", "WorkflowStep"]

Block = Union[Actions, Context, Divider, Header, ImageBlock, Input, Section]


class View(Component):
    blocks: List[Block] = Field(..., min_items=1, max_items=100)
    private_metadata: Optional[str] = Field(None, min_length=1, max_length=3000)
    callback_id: Optional[str] = Field(None, min_length=1, max_length=255)


class Home(View):
    type: str = "home"
    external_id: Optional[str] = Field(None, min_length=1, max_length=255)

    def __init__(
        self,
        *,
        blocks: List[Block],
        private_metadata: Optional[str] = None,
        callback_id: Optional[str] = None,
        external_id: Optional[str] = None,
    ):
        super().__init__(
            blocks=blocks,
            private_metadata=private_metadata,
            callback_id=callback_id,
            external_id=external_id,
        )


class Modal(View):
    type: str = "modal"
    title: PlainText
    close: Optional[PlainText] = None
    submit: Optional[PlainText] = None
    clear_on_close: Optional[bool] = None
    notify_on_close: Optional[bool] = None
    external_id: Optional[str] = Field(None, min_length=1, max_length=255)
    submit_disabled: Optional[bool] = None

    def __init__(
        self,
        *,
        title: PlainText,
        blocks: List[Block],
        close: Optional[PlainText] = None,
        submit: Optional[PlainText] = None,
        private_metadata: Optional[str] = None,
        callback_id: Optional[str] = None,
        clear_on_close: Optional[bool] = None,
        notify_on_close: Optional[bool] = None,
        external_id: Optional[str] = None,
        submit_disabled: Optional[bool] = None,
    ):
        super(View, self).__init__(
            title=title,
            blocks=blocks,
            close=close,
            submit=submit,
            private_metadata=private_metadata,
            callback_id=callback_id,
            clear_on_close=clear_on_close,
            notify_on_close=notify_on_close,
            external_id=external_id,
            submit_disabled=submit_disabled,
        )

    _validate_title = validator("title", validate_text_length, max_len=24)
    _validate_close = validator("close", validate_text_length, max_len=24)
    _validate_submit = validator("submit", validate_text_length, max_len=24)

    @root_validator
    def _validate_values(cls, values):
        blocks = values.get("blocks")
        submit = values.get("submit")

        if blocks and not submit and Input in (type(b) for b in values["blocks"]):
            raise ValueError("submit is required when an Input is within blocks")
        return values


class WorkflowStep(View):
    type: str = "workflow_step"
    submit_disabled: Optional[bool] = None

    def __init__(
        self,
        *,
        blocks: List[Block],
        private_metadata: Optional[str] = None,
        callback_id: Optional[str] = None,
        submit_disabled: Optional[bool] = None,
    ):
        super(View, self).__init__(
            blocks=blocks,
            private_metadata=private_metadata,
            callback_id=callback_id,
            submit_disabled=submit_disabled,
        )


class Message(Component):
    blocks: List[Block] = Field(..., min_items=1, max_items=50)

    def __init__(self, *, blocks: List[Block]):
        super().__init__(blocks=blocks)



import itertools
from datetime import date, time
from typing import List, Literal, Optional

from pydantic import Field, root_validator
from pydantic.networks import HttpUrl

from blockkit.components import Component
from blockkit.objects import (
    Confirm,
    DispatchActionConfig,
    Filter,
    Option,
    OptionGroup,
    PlainText,
)
from blockkit.validators import (
    SlackUrl,
    validate_date,
    validate_text_length,
    validate_time,
    validator,
)

__all__ = [
    "Button",
    "ChannelsSelect",
    "Checkboxes",
    "ConversationsSelect",
    "DatePicker",
    "ExternalSelect",
    "Image",
    "MultiChannelsSelect",
    "MultiConversationsSelect",
    "MultiExternalSelect",
    "MultiStaticSelect",
    "MultiUsersSelect",
    "Overflow",
    "PlainTextInput",
    "RadioButtons",
    "StaticSelect",
    "Timepicker",
    "UsersSelect",
]


class ActionableComponent(Component):
    action_id: Optional[str] = Field(None, min_length=1, max_length=255)


class Button(ActionableComponent):
    type: str = "button"
    text: PlainText
    url: Optional[SlackUrl] = None
    value: Optional[str] = Field(None, min_length=1, max_length=2000)
    style: Optional[Literal["primary", "danger"]] = None
    confirm: Optional[Confirm] = None

    def __init__(
        self,
        *,
        text: PlainText,
        action_id: Optional[str] = None,
        url: Optional[SlackUrl] = None,
        value: Optional[str] = None,
        style: Optional[Literal["primary", "danger"]] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            text=text,
            action_id=action_id,
            url=url,
            value=value,
            style=style,
            confirm=confirm,
        )

    _validate_text = validator("text", validate_text_length, max_len=75)


class Checkboxes(ActionableComponent):
    type: str = "checkboxes"
    options: List[Option] = Field(..., min_items=1, max_items=10)
    initial_options: Optional[List[Option]] = Field(None, min_items=1, max_items=10)
    confirm: Optional[Confirm] = None

    def __init__(
        self,
        *,
        options: List[Option],
        action_id: Optional[str] = None,
        initial_options: Optional[List[Option]] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            options=options,
            action_id=action_id,
            initial_options=initial_options,
            confirm=confirm,
        )

    @root_validator
    def _validate_values(cls, values):
        initial_options = values.get("initial_options")
        options = values.get("options")

        if initial_options is not None:
            for initial_option in initial_options:
                if initial_option not in options:
                    raise ValueError(f"Option {initial_option} isn't within {options}")
        return values


class DatePicker(ActionableComponent):
    type: str = "datepicker"
    placeholder: Optional[PlainText] = None
    initial_date: Optional[date] = None
    confirm: Optional[Confirm] = None

    def __init__(
        self,
        *,
        action_id: Optional[str] = None,
        placeholder: Optional[PlainText] = None,
        initial_date: Optional[date] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            action_id=action_id,
            placeholder=placeholder,
            initial_date=initial_date,
            confirm=confirm,
        )

    _validate_placeholder = validator("placeholder", validate_text_length, max_len=150)
    _validate_initial_date = validator("initial_date", validate_date)


class Image(Component):
    type: str = "image"
    image_url: HttpUrl
    alt_text: str = Field(..., min_length=1, max_length=2000)

    def __init__(self, *, image_url: HttpUrl, alt_text: str):
        super().__init__(image_url=image_url, alt_text=alt_text)


class Select(ActionableComponent):
    placeholder: PlainText
    confirm: Optional[Confirm] = None

    _validate_placeholder = validator("placeholder", validate_text_length, max_len=150)


class StaticSelectBase(Select):
    options: Optional[List[Option]] = Field(None, min_items=1, max_items=100)
    option_groups: Optional[List[OptionGroup]] = Field(None, min_items=1, max_items=100)


class StaticSelect(StaticSelectBase):
    type: str = "static_select"
    initial_option: Optional[Option] = None

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        options: Optional[List[Option]] = None,
        option_groups: Optional[List[OptionGroup]] = None,
        initial_option: Optional[Option] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            options=options,
            option_groups=option_groups,
            initial_option=initial_option,
            confirm=confirm,
        )

    @root_validator
    def _validate_values(cls, values):
        initial_option = values.get("initial_option")
        options = values.get("options")
        option_groups = values.get("option_groups")

        if not any((options, option_groups)) or options and option_groups:
            raise ValueError("You must provide either options or option_groups")

        if None not in (initial_option, options) and initial_option not in options:
            raise ValueError(f"Option {initial_option} isn't within {options}")

        if None not in (initial_option, option_groups):
            if initial_option not in itertools.chain(
                *[og.options for og in option_groups]
            ):
                raise ValueError(
                    f"Option {initial_option} isn't within {option_groups}"
                )

        return values


class MultiStaticSelect(StaticSelectBase):
    type: str = "multi_static_select"
    initial_options: Optional[List[Option]] = None
    max_selected_items: Optional[int] = Field(..., gt=0)

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        options: Optional[List[Option]] = None,
        option_groups: Optional[List[OptionGroup]] = None,
        initial_options: Optional[Option] = None,
        confirm: Optional[Confirm] = None,
        max_selected_items: Optional[int] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            options=options,
            option_groups=option_groups,
            initial_options=initial_options,
            confirm=confirm,
            max_selected_items=max_selected_items,
        )

    @root_validator
    def _validate_values(cls, values):
        initial_options = values.get("initial_options")
        options = values.get("options")
        option_groups = values.get("option_groups")

        if not any((options, option_groups)) or options and option_groups:
            raise ValueError("You must provide either options or option_groups")

        if None not in (initial_options, options):
            for initial_option in initial_options:
                if initial_option not in options:
                    raise ValueError(f"Option {initial_option} isn't within {options}")

        if None not in (initial_options, option_groups):
            groups_options = itertools.chain(*[og.options for og in option_groups])
            for initial_option in initial_options:
                if initial_option not in groups_options:
                    raise ValueError(
                        f"Option {initial_option} isn't within {option_groups}"
                    )

        return values


class ExternalSelectBase(Select):
    min_query_length: Optional[int] = Field(None, gt=0)


class ExternalSelect(ExternalSelectBase):
    type: str = "external_select"
    initial_option: Optional[Option] = None

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_option: Optional[Option] = None,
        min_query_length: Optional[int] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_option=initial_option,
            min_query_length=min_query_length,
            confirm=confirm,
        )


class MultiExternalSelect(ExternalSelectBase):
    type: str = "multi_external_select"
    initial_options: Optional[List[Option]] = None
    max_selected_items: Optional[int] = Field(None, gt=0)

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        min_query_length: Optional[int] = None,
        initial_options: Optional[List[Option]] = None,
        confirm: Optional[Confirm] = None,
        max_selected_items: Optional[int] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            min_query_length=min_query_length,
            initial_options=initial_options,
            confirm=confirm,
            max_selected_items=max_selected_items,
        )


class UsersSelect(Select):
    type: str = "users_select"
    initial_user: Optional[str] = Field(None, min_length=1)

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_user: Optional[str] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_user=initial_user,
            confirm=confirm,
        )


class MultiUsersSelect(Select):
    type: str = "multi_users_select"
    initial_users: Optional[List[str]] = None
    max_selected_items: Optional[int] = Field(None, gt=0)

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_users: Optional[str] = None,
        confirm: Optional[Confirm] = None,
        max_selected_items: Optional[int] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_users=initial_users,
            confirm=confirm,
            max_selected_items=max_selected_items,
        )


class ConversationsSelect(Select):
    type: str = "conversations_select"
    initial_conversation: Optional[str] = Field(None, min_length=1)
    default_to_current_conversation: Optional[bool] = None
    response_url_enabled: Optional[bool] = None
    filter: Optional[Filter] = None

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_conversation: Optional[str] = None,
        default_to_current_conversation: Optional[bool] = None,
        confirm: Optional[Confirm] = None,
        response_url_enabled: Optional[bool] = None,
        filter: Optional[Filter] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_conversation=initial_conversation,
            default_to_current_conversation=default_to_current_conversation,
            confirm=confirm,
            response_url_enabled=response_url_enabled,
            filter=filter,
        )


class MultiConversationsSelect(Select):
    type: str = "multi_conversations_select"
    initial_conversations: Optional[List[str]] = None
    default_to_current_conversation: Optional[bool] = None
    max_selected_items: Optional[int] = Field(None, gt=0)
    filter: Optional[Filter] = None

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_conversations: Optional[List[str]] = None,
        default_to_current_conversation: Optional[bool] = None,
        confirm: Optional[Confirm] = None,
        max_selected_items: Optional[int] = None,
        filter: Optional[Filter] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_conversations=initial_conversations,
            default_to_current_conversation=default_to_current_conversation,
            confirm=confirm,
            max_selected_items=max_selected_items,
            filter=filter,
        )


class ChannelsSelect(Select):
    type: str = "channels_select"
    initial_channel: Optional[str] = Field(None, min_length=1)
    response_url_enabled: Optional[bool] = None

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_channel: Optional[str] = None,
        confirm: Optional[Confirm] = None,
        response_url_enabled: Optional[bool] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_channel=initial_channel,
            confirm=confirm,
            response_url_enabled=response_url_enabled,
        )


class MultiChannelsSelect(Select):
    type: str = "multi_channels_select"
    initial_channels: Optional[List[str]] = None
    max_selected_items: Optional[int] = Field(None, gt=0)

    def __init__(
        self,
        *,
        placeholder: PlainText,
        action_id: Optional[str] = None,
        initial_channels: Optional[str] = None,
        confirm: Optional[Confirm] = None,
        max_selected_items: Optional[int] = None,
    ):
        super().__init__(
            placeholder=placeholder,
            action_id=action_id,
            initial_channels=initial_channels,
            confirm=confirm,
            max_selected_items=max_selected_items,
        )


class Overflow(ActionableComponent):
    type: str = "overflow"
    options: List[Option] = Field(..., min_items=2, max_items=5)
    confirm: Optional[Confirm] = None

    def __init__(
        self,
        *,
        action_id: Optional[str] = None,
        options: List[Option],
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(action_id=action_id, options=options, confirm=confirm)


class PlainTextInput(ActionableComponent):
    type: str = "plain_text_input"
    placeholder: Optional[PlainText] = None
    initial_value: Optional[str] = Field(None, min_length=1)
    multiline: Optional[bool] = None
    min_length: Optional[int] = Field(None, ge=0, lt=3000)
    max_length: Optional[int] = Field(None, ge=0, lt=3000)
    dispatch_action_config: Optional[DispatchActionConfig] = None

    def __init__(
        self,
        *,
        action_id: Optional[str] = None,
        placeholder: Optional[PlainText] = None,
        initial_value: Optional[str] = None,
        multiline: Optional[bool] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        dispatch_action_config: Optional[DispatchActionConfig] = None,
    ):
        super().__init__(
            action_id=action_id,
            placeholder=placeholder,
            initial_value=initial_value,
            multiline=multiline,
            min_length=min_length,
            max_length=max_length,
            dispatch_action_config=dispatch_action_config,
        )

    _validate_placeholder = validator("placeholder", validate_text_length, max_len=150)


class RadioButtons(ActionableComponent):
    type: str = "radio_buttons"
    options: List[Option] = Field(..., min_items=1, max_items=10)
    initial_option: Optional[Option] = None
    confirm: Optional[Confirm] = None

    def __init__(
        self,
        *,
        options: List[Option],
        action_id: Optional[str] = None,
        initial_option: Optional[Option] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            options=options,
            action_id=action_id,
            initial_option=initial_option,
            confirm=confirm,
        )

    @root_validator
    def _validate_values(cls, values):
        initial_option = values.get("initial_option")
        options = values.get("options")

        if initial_option is not None and initial_option not in options:
            raise ValueError(f"Option {initial_option} isn't within {options}")
        return values


class Timepicker(ActionableComponent):
    type: str = "timepicker"
    placeholder: Optional[PlainText] = None
    initial_time: Optional[time] = None
    confirm: Optional[Confirm] = None

    _validate_placeholder = validator("placeholder", validate_text_length, max_len=150)
    _validate_initial_time = validator("initial_time", validate_time)

    def __init__(
        self,
        *,
        action_id: Optional[str] = None,
        placeholder: Optional[PlainText] = None,
        initial_time: Optional[time] = None,
        confirm: Optional[Confirm] = None,
    ):
        super().__init__(
            action_id=action_id,
            placeholder=placeholder,
            initial_time=initial_time,
            confirm=confirm,
        )



from pydantic import AnyUrl
from pydantic import validator as pydantic_validator


def validator(field, func, **kwargs):
    return pydantic_validator(field, allow_reuse=True)(lambda v: func(v, **kwargs))


def validate_text_length(v, *, max_len):
    if v is not None and len(v.text) > max_len:
        raise ValueError(f"Maximum length is {max_len} characters")
    return v


def validate_list_text_length(v, *, max_len):
    if v is None:
        return
    for item in v:
        validate_text_length(item, max_len=max_len)
    return v


def validate_date(v):
    if v is not None:
        return v.isoformat()
    return v


def validate_time(v):
    if v is not None:
        return v.strftime("%H:%M")
    return v


class SlackUrl(AnyUrl):
    max_length = 3000



from typing import List, Optional, Union

from pydantic import Field
from pydantic.class_validators import root_validator

from blockkit.components import Component
from blockkit.enums import Include, Style, TriggerActionsOn
from blockkit.validators import SlackUrl, validate_text_length, validator

__all__ = [
    "Confirm",
    "DispatchActionConfig",
    "Filter",
    "MarkdownText",
    "Option",
    "OptionGroup",
    "PlainText",
]


class MarkdownText(Component):
    type: str = "mrkdwn"
    text: str = Field(..., min_length=1)
    verbatim: Optional[bool] = None

    def __init__(self, *, text: str, verbatim: Optional[bool] = None):
        super().__init__(text=text, verbatim=verbatim)


class PlainText(Component):
    type: str = "plain_text"
    text: str = Field(..., min_length=1)
    emoji: Optional[bool] = None

    def __init__(self, *, text: str, emoji: Optional[bool] = None):
        super().__init__(text=text, emoji=emoji)


class Confirm(Component):
    title: PlainText
    text: Union[PlainText, MarkdownText]
    confirm: PlainText
    deny: PlainText
    style: Optional[Style] = None

    def __init__(
        self,
        *,
        title: PlainText,
        text: Union[PlainText, MarkdownText],
        confirm: PlainText,
        deny: PlainText,
        style: Optional[Style] = None,
    ):
        super().__init__(
            title=title, text=text, confirm=confirm, deny=deny, style=style
        )

    _validate_title = validator("title", validate_text_length, max_len=100)
    _validate_text = validator("text", validate_text_length, max_len=300)
    _validate_confirm = validator("confirm", validate_text_length, max_len=30)
    _validate_deny = validator("deny", validate_text_length, max_len=30)


class Option(Component):
    text: Union[PlainText, MarkdownText]
    value: str = Field(..., min_length=1, max_length=75)
    description: Optional[PlainText] = None
    url: Optional[SlackUrl] = None

    def __init__(
        self,
        *,
        text: Union[PlainText, MarkdownText],
        value: str,
        description: Optional[PlainText] = None,
        url: Optional[SlackUrl] = None,
    ):
        super().__init__(text=text, value=value, description=description, url=url)

    _validate_text = validator("text", validate_text_length, max_len=75)
    _validate_description = validator("description", validate_text_length, max_len=75)


class OptionGroup(Component):
    label: PlainText
    options: List[Option] = Field(..., min_items=1, max_items=100)

    def __init__(self, *, label: PlainText, options: List[Option]):
        super().__init__(label=label, options=options)

    _validate_label = validator("label", validate_text_length, max_len=75)


class DispatchActionConfig(Component):
    trigger_actions_on: List[TriggerActionsOn] = Field(..., min_items=1, max_items=2)

    def __init__(self, *, trigger_actions_on: List[TriggerActionsOn]):
        super().__init__(trigger_actions_on=trigger_actions_on)


class Filter(Component):
    include: Optional[List[Include]] = None
    exclude_external_shared_channels: Optional[bool] = None
    exclude_bot_users: Optional[bool] = None

    def __init__(
        self,
        *,
        include: Optional[List[Include]] = None,
        exclude_external_shared_channels: Optional[bool] = None,
        exclude_bot_users: Optional[bool] = None,
    ):
        super().__init__(
            include=include,
            exclude_external_shared_channels=exclude_external_shared_channels,
            exclude_bot_users=exclude_bot_users,
        )

    @root_validator
    def _validate_values(cls, values):
        if not any(values.values()):
            raise ValueError("You should provide at least one argument.")
        return values



from typing import List, Optional, Union

from pydantic import Field, root_validator
from pydantic.networks import HttpUrl

from blockkit.components import Component
from blockkit.elements import (
    Button,
    ChannelsSelect,
    Checkboxes,
    ConversationsSelect,
    DatePicker,
    ExternalSelect,
    Image,
    MultiChannelsSelect,
    MultiConversationsSelect,
    MultiExternalSelect,
    MultiStaticSelect,
    MultiUsersSelect,
    Overflow,
    PlainTextInput,
    RadioButtons,
    StaticSelect,
    Timepicker,
    UsersSelect,
)
from blockkit.objects import MarkdownText, PlainText
from blockkit.validators import (
    validate_list_text_length,
    validate_text_length,
    validator,
)

__all__ = ["Actions", "Context", "Divider", "Header", "ImageBlock", "Input", "Section"]

ActionElement = Union[
    Button,
    Checkboxes,
    DatePicker,
    StaticSelect,
    MultiStaticSelect,
    ExternalSelect,
    MultiExternalSelect,
    UsersSelect,
    MultiUsersSelect,
    ConversationsSelect,
    MultiConversationsSelect,
    ChannelsSelect,
    MultiChannelsSelect,
    Overflow,
    RadioButtons,
    Timepicker,
]


class Block(Component):
    block_id: Optional[str] = Field(None, min_length=1, max_length=255)


class Actions(Block):
    type: str = "actions"
    elements: List[ActionElement] = Field(..., min_items=1, max_items=5)

    def __init__(
        self, *, elements: List[ActionElement], block_id: Optional[str] = None
    ):
        super().__init__(elements=elements, block_id=block_id)


class Context(Block):
    type: str = "context"
    elements: List[Union[Image, PlainText, MarkdownText]] = Field(
        ..., min_items=1, max_items=10
    )

    def __init__(
        self,
        *,
        elements: List[Union[Image, PlainText, MarkdownText]],
        block_id: Optional[str] = None,
    ):
        super().__init__(elements=elements, block_id=block_id)


class Divider(Block):
    type: str = "divider"

    def __init__(self, *, block_id: Optional[str] = None):
        super().__init__(block_id=block_id)


class Header(Block):
    type: str = "header"
    text: PlainText

    def __init__(self, *, text: PlainText, block_id: Optional[str] = None):
        super().__init__(text=text, block_id=block_id)

    _validate_text = validator("text", validate_text_length, max_len=150)


class ImageBlock(Block):
    type: str = "image"
    image_url: HttpUrl
    alt_text: str = Field(..., min_length=1, max_length=2000)
    title: Optional[PlainText] = None

    def __init__(
        self,
        *,
        image_url: HttpUrl,
        alt_text: str,
        title: Optional[PlainText] = None,
        block_id: Optional[str] = None,
    ):
        super().__init__(
            image_url=image_url, alt_text=alt_text, title=title, block_id=block_id
        )

    _validate_title = validator("title", validate_text_length, max_len=2000)


InputElement = Union[
    PlainTextInput,
    Checkboxes,
    RadioButtons,
    StaticSelect,
    MultiStaticSelect,
    ExternalSelect,
    MultiExternalSelect,
    UsersSelect,
    MultiUsersSelect,
    ConversationsSelect,
    MultiConversationsSelect,
    ChannelsSelect,
    MultiChannelsSelect,
    DatePicker,
]


class Input(Block):
    type: str = "input"
    label: PlainText
    element: InputElement
    dispatch_action: Optional[bool] = None
    hint: Optional[PlainText] = None
    optional: Optional[bool] = None

    def __init__(
        self,
        *,
        label: PlainText,
        element: InputElement,
        block_id: Optional[str] = None,
        dispatch_action: Optional[bool] = None,
        hint: Optional[PlainText] = None,
        optional: Optional[bool] = None,
    ):
        super().__init__(
            label=label,
            element=element,
            block_id=block_id,
            dispatch_action=dispatch_action,
            hint=hint,
            optional=optional,
        )

    _validate_label = validator("label", validate_text_length, max_len=2000)
    _validate_hint = validator("hint", validate_text_length, max_len=2000)


AccessoryElement = Union[
    Button,
    Checkboxes,
    DatePicker,
    Image,
    StaticSelect,
    MultiStaticSelect,
    ExternalSelect,
    MultiExternalSelect,
    UsersSelect,
    MultiUsersSelect,
    ConversationsSelect,
    MultiConversationsSelect,
    ChannelsSelect,
    MultiChannelsSelect,
    Overflow,
    RadioButtons,
    Timepicker,
]


class Section(Block):
    type: str = "section"
    text: Optional[Union[PlainText, MarkdownText]] = None
    fields_: Optional[List[Union[PlainText, MarkdownText]]] = Field(
        None, alias="fields", min_items=1, max_items=10
    )
    accessory: Optional[AccessoryElement] = None

    def __init__(
        self,
        *,
        text: Optional[Union[PlainText, MarkdownText]] = None,
        block_id: Optional[str] = None,
        fields: Optional[List[Union[PlainText, MarkdownText]]] = None,
        accessory: Optional[AccessoryElement] = None,
    ):
        super().__init__(
            text=text, block_id=block_id, fields=fields, accessory=accessory
        )

    _validate_text = validator("text", validate_text_length, max_len=3000)
    _validate_fields = validator("fields_", validate_list_text_length, max_len=2000)

    @root_validator
    def _validate_values(cls, values):
        text = values.get("text")
        fields = values.get("fields_")

        if text is None and fields is None:
            raise ValueError("You must provide either text or fields")

        return values



from black import FileMode, format_str

from blockkit.blocks import (
    Actions,
    Context,
    Divider,
    Header,
    ImageBlock,
    Input,
    Section,
)
from blockkit.elements import (
    Button,
    ChannelsSelect,
    Checkboxes,
    ConversationsSelect,
    DatePicker,
    ExternalSelect,
    Image,
    MultiChannelsSelect,
    MultiConversationsSelect,
    MultiExternalSelect,
    MultiStaticSelect,
    MultiUsersSelect,
    Overflow,
    PlainTextInput,
    RadioButtons,
    StaticSelect,
    UsersSelect,
)
from blockkit.objects import (
    Confirm,
    DispatchActionConfig,
    Filter,
    MarkdownText,
    Option,
    OptionGroup,
    PlainText,
)
from blockkit.surfaces import Home, Message, Modal


class CodeGenerationError(Exception):
    pass


components = {
    "section": Section,
    "plain_text": PlainText,
    "mrkdwn": MarkdownText,
    "button": Button,
    "static_select": StaticSelect,
    "users_select": UsersSelect,
    "channels_select": ChannelsSelect,
    "conversations_select": ConversationsSelect,
    "external_select": ExternalSelect,
    "multi_static_select": MultiStaticSelect,
    "multi_users_select": MultiUsersSelect,
    "multi_channels_select": MultiChannelsSelect,
    "multi_conversations_select": MultiConversationsSelect,
    "multi_external_select": MultiExternalSelect,
    "image": ImageBlock,
    "overflow": Overflow,
    "datepicker": DatePicker,
    "checkboxes": Checkboxes,
    "radio_buttons": RadioButtons,
    "actions": Actions,
    "divider": Divider,
    "context": Context,
    "input": Input,
    "plain_text_input": PlainTextInput,
    "header": Header,
    "modal": Modal,
    "home": Home,
}


def generate(payload):
    code = _generate(payload)
    eval_components(code)
    return code


def generate_pretty(payload):
    return format_str(generate(payload), mode=FileMode())


def _generate(payload, component=None):
    kwargs = []
    for name, value in payload.items():
        if name == "type":
            continue

        if type(value) == str:
            if name == "text":
                value = (
                    value.replace("\n", "\\n").replace("\t", "\\t").replace('"', '\\"')
                )
            kwarg = f'{name}="{value}"'
        elif type(value) == dict:
            subcomponent = None
            if name == "filter":
                subcomponent = Filter
            elif name == "dispatch_action_config":
                subcomponent = DispatchActionConfig
            elif name == "confirm" and "confirm" in value:
                subcomponent = Confirm
            elif name == "accessory" and value.get("type") == "image":
                subcomponent = Image
            kwarg = f"{name}=" + _generate(value, subcomponent)
        elif type(value) == list:
            if name == "options":
                items = [_generate(v, Option) for v in value]
            elif name == "option_groups":
                items = [_generate(v, OptionGroup) for v in value]
            elif name in ("include", "trigger_actions_on"):
                items = [f'"{v}"' for v in value]
            elif name == "elements":
                items = [
                    _generate(v, Image) if v.get("type") == "image" else _generate(v)
                    for v in value
                ]
            else:
                items = [_generate(v) for v in value]

            items = ", ".join(items)
            kwarg = f"{name}=[{items}]"
        else:
            kwarg = f"{name}={value}"

        kwargs.append(kwarg)

    if not component:
        if "blocks" in payload and "type" not in payload:
            component = Message
        else:
            type_ = payload.get("type")
            try:
                component = components[type_]
            except KeyError:
                raise CodeGenerationError(
                    f'Can\'t _generate component of type "{type_}."'
                )

    class_name = component.__name__
    kwargs = ", ".join(kwargs)

    return f"{class_name}({kwargs})"


allowed_names = {c.__name__: c for c in components.values()}
allowed_names.update(
    {
        c.__name__: c
        for c in (
            Message,
            Filter,
            DispatchActionConfig,
            Confirm,
            Image,
            Option,
            OptionGroup,
        )
    }
)


def eval_components(code):
    compiled_code = compile(code, "<string>", "eval")
    for name in compiled_code.co_names:
        if name not in allowed_names:
            raise NameError(f"Use of {name} not allowed")
    return eval(code, {"__builtins__": {}}, allowed_names)
